"""Execution of scheduled tasks."""
from __future__ import annotations

import datetime
import os
import shutil
import signal
import subprocess
import sys
import time
import typing

from descriptors import cachedproperty, classonlymethod

from fate.common.log import LogReader
from fate.common.output import CompletingTask
from fate.util import stream

from .. import ext

from .base import InvokedTask
from .event import TaskEvents, TaskLogEvent, TaskReadyEvent
from .failed_task import TaskInvocationError


class PipeRW(typing.NamedTuple):
    """A readable output and writable input pair of file descriptors.

    Though seemingly the same as a single pipe -- with a readable end
    and a writable end -- this structure is intended to collect *one*
    end of a *pair* of pipes.

    """
    output: int
    input: int


class Pipe(typing.NamedTuple):
    """An OS pipe consisting of a readable output end and a writable
    input end.

    """
    output: int
    input: int

    @classonlymethod
    def open(cls):
        """Create and construct a Pipe."""
        return cls._make(os.pipe())


class _TaskProcess(typing.NamedTuple):

    process: subprocess.Popen
    stdin: stream.BufferedInput
    statein: stream.BufferedInput
    stdout: stream.ProgressiveOutput
    stderr: stream.BufferedOutput
    stateout: stream.BufferedOutput


class SpawnedTask(ext.BoundTask, InvokedTask, CompletingTask):
    """Task whose process has been spawned."""

    #
    # state communication pipes
    #
    # we'll ensure that our state pipes are available (copied) to descriptors
    # 3 & 4 in the child process (for simplicity)
    #
    _state_child_ = PipeRW(input=3, output=4)

    #
    # in the parent process, each task's pipes will be provisioned once
    # (and file descriptors cached)
    #
    _statein_ = cachedproperty.static(Pipe.open)

    _stateout_ = cachedproperty.static(Pipe.open)

    @staticmethod
    def _dup_fd_(src, dst):
        """Duplicate (copy) file descriptor `src` to `dst`.

        `dst` *may not* be one of the standard file descriptors (0-2).
        `dst` is not otherwise checked.

        The duplicate descriptor is set inheritable.

        It is presumed that this method is used in the context of a
        process fork, *e.g.* as the `preexec_fn` of `subprocess.Popen`
        -- and with `close_fds=True`. (As such, any file descriptor may
        be available for use as `dst`.)

        """
        if src == dst:
            return

        if dst < 3:
            raise ValueError(f"will not overwrite standard file descriptor: {dst}")

        os.dup2(src, dst, inheritable=True)

    def _set_fds_(self):
        """Duplicate inherited state file descriptors to conventional
        values in the task subprocess.

        """
        for (parent, child) in zip(self._state_parent_, reversed(self._state_child_)):
            self._dup_fd_(parent, child)

    @property
    def _state_parent_(self):
        """The parent process's originals of its child's pair of
        readable and writable state file descriptors.

        """
        return PipeRW(self._statein_.output, self._stateout_.input)

    @property
    def _pass_fds_(self):
        """The child process's readable and writable state file
        descriptors -- *both* the originals and their desired
        conventional values.

        These descriptors must be inherited by the child process -- and
        not closed -- for inter-process communication of task state.

        """
        return self._state_parent_ + self._state_child_

    @cachedproperty
    def _stateinfile_(self) -> typing.BinaryIO:
        return open(self._statein_.input, 'wb')

    @cachedproperty
    def _stateoutfile_(self) -> typing.BinaryIO:
        return open(self._stateout_.output, 'rb')

    @InvokedTask._constructor_
    def spawn(cls, task, state):
        """Construct a SpawnedTask extending the specified Task."""
        spawned = cls(task, state)

        # _constructor_ would otherwise handle linking/adoption for us;
        # but, we need this in order to spawn, so we'll do it here:
        cls._link_(spawned, task)

        spawned._spawn_()

        return spawned

    def __init__(self, data, /, state):
        super().__init__(data, state)

        self._process_ = None
        self._started_ = None
        self._ended_ = None

        self.terminated_ = None
        self.killed_ = None

        self.stdout_ = None
        self.stderr_ = None
        self.stdin_ = None
        self.statein_ = None
        self.stateout_ = None

        self._events_ = None
        self._log_reader = None

    @property
    def stopped_(self) -> typing.Optional[float]:
        return self.killed_ or self.terminated_

    def _spawn_(self):
        if self._process_ is not None:
            raise ValueError("task already spawned")

        (
            self._process_,
            self.stdin_,
            self.statein_,
            self.stdout_,
            self.stderr_,
            self.stateout_,
        ) = self._popen()

        self._started_ = time.time()

        self._events_ = TaskEvents()
        self._log_reader = LogReader(self.stderr_)

    def _preexec_legacy(self) -> None:
        # Assign (duplicate) state file descriptors to expected values
        self._set_fds_()

        # Assign the child process its own new process group
        os.setpgrp()

    def _popen(self) -> _TaskProcess:
        (program, *args) = self.exec_

        executable = shutil.which(program)

        if executable is None:
            raise TaskInvocationError(f'command not found on path: {program}')

        # We prefer to have Popen handle child setup as much as possible
        # so we opt into new features as they become available.
        if sys.version_info < (3, 11):
            # Popen doesn't yet offer process_group so we'll do it ourselves
            kwargs = dict(
                preexec_fn=self._preexec_legacy,
            )
        else:
            # We can just use process_group
            kwargs = dict(
                # Assign the child process its own new process group
                process_group=0,

                preexec_fn=self._set_fds_,
            )

        process = subprocess.Popen(
            [executable] + args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            pass_fds=self._pass_fds_,
            **kwargs,
        )

        result = _TaskProcess(
            process=process,

            # stdout needn't be inspected until the task completes; and, synchronous, non-blocking
            # processing of the pipe is relatively inefficient (for large payloads). instead,
            # we'll launch a daemon thread to sit on it and read it as efficiently as possible.
            stdout=stream.progressive_output(process.stdout,
                                             f'Reader ({self.__name__} {process.pid}): stdout'),

            # we don't expect any other IPC data to be huge; and, at least in the case of
            # stderr, we want to inspect it as it comes in.
            #
            # for simplicity: make pipe descriptors non-blocking & initialize buffer handlers
            #
            # (note: this works for pipes on Win32 but only as of Py312)
            stderr=stream.nonblocking_output(process.stderr),
            stateout=stream.nonblocking_output(self._stateoutfile_),

            stdin=stream.nonblocking_input(process.stdin, self.param_.encode()),
            statein=stream.nonblocking_input(self._stateinfile_, self._state_.read().encode()),
        )

        # write inputs (at least up to buffer size)
        result.stdin.send()
        result.statein.send()

        # close child's descriptors in parent process
        for parent_desc in self._state_parent_:
            os.close(parent_desc)

        return result

    def started_(self) -> typing.Optional[float]:
        return self._started_

    def ended_(self) -> typing.Optional[float]:
        if self._ended_ is None:
            self.poll_()

        return self._ended_

    def duration_(self) -> typing.Optional[datetime.timedelta]:
        return (ended := self.ended_()) and datetime.timedelta(seconds=ended - self.started_())

    def expires_(self) -> typing.Optional[float]:
        if (started := self.started_()) is None:
            return None

        if (timeout := self.timeout_) is None:
            return None

        return started + timeout.total_seconds()

    def expired_(self) -> bool:
        expires = self.expires_()
        return expires is not None and expires <= time.time()

    def _signal(self, signal) -> None:
        if self._process_ is None:
            raise ValueError("task not spawned")

        if os.getpgid(self._process_.pid) == self._process_.pid:
            # as expected: signal group
            os.killpg(self._process_.pid, signal)
        else:
            # unexpected: stick to process itself
            self._process_.send_signal(signal)

    def _terminate_(self) -> None:
        self._signal(signal.SIGTERM)
        self.terminated_ = time.time()

    def _kill_(self) -> None:
        self._signal(signal.SIGKILL)
        self.killed_ = time.time()

    def events_(self) -> typing.Optional[TaskEvents]:
        if self._events_ is not None and not self._events_.closed:
            self.poll_()

        return self._events_

    def _record_events_(self, returncode: typing.Optional[int]) -> None:
        if self._events_ is None:
            raise ValueError("task not spawned")

        if self._events_.closed:
            return

        final = returncode is not None

        for record in self._log_reader.read(final):
            self._events_.write(TaskLogEvent(self, record))

        if final:
            self._events_.write(TaskReadyEvent(self, returncode))
            self._events_.close()

    def poll_(self) -> typing.Optional[int]:
        """Check whether the task program has exited and return its exit
        code if any.

        If the task has expired, i.e. run past a configured timeout,
        this method sends the process the TERM signal; if execution has
        continued past this, the KILL signal is sent.

        BufferedInput and BufferedOutput handlers are invoked to send/
        receive remaining data.

        Sets the SpawnedTask's `ended` time, and records the task's
        state output, when the process has terminated.

        """
        if self.expired_() and self._process_.poll() is None:
            if self.terminated_ is None:
                self._terminate_()
            else:
                self._kill_()

        returncode = self._process_.poll()

        self.stdin_.send()
        self.statein_.send()

        self.stderr_.receive()
        self.stateout_.receive()

        if returncode is not None and self._ended_ is None:
            self._ended_ = time.time()

            self.stdout_.close()

            # Note: with retry this will also permit 42
            if returncode == 0:
                self._state_.write(str(self.stateout_))

        self._record_events_(returncode)

        return returncode

    def ready_(self) -> bool:
        """Return whether the task program's process has terminated.

        See poll_().

        """
        return self.poll_() is not None



"""Scheduling of tasks for execution."""
from . import ext, invoked_task as invocation


class ScheduledTask(ext.BoundTask):
    """Task extended for scheduling to be invoked."""

    @ext.BoundTask._constructor_
    def schedule(cls, task, state):
        """Construct a ScheduledTask extending the specified Task."""
        return cls(task, state)

    def __call__(self) -> invocation.InvokedTask:
        """Execute the task's program in a background process.

        Returns an InvokedTask – either a successfully-spawned
        SpawnedTask or a FailedInvocationTask.

        """
        try:
            return invocation.SpawnedTask.spawn(self, self._state_)
        except invocation.TaskInvocationError as exc:
            return invocation.FailedInvocationTask.fail(self, exc)



from __future__ import annotations

import logging
import re
import typing
from dataclasses import dataclass, field

from fate.conf.error import ConfValueError, LogRecordDecodeError
from fate.util.format import SLoader

if typing.TYPE_CHECKING:
    from fate.util.stream import BufferedOutput


@dataclass
class LogReader:

    stream: typing.Union[bytes, BufferedOutput]
    position: int = field(default=0, init=False)

    @property
    def buffer(self) -> bytes:
        return bytes(self.stream)[self.position:]

    def read(self, final=False):
        position = self.position

        for match in re.finditer(rb'(?P<record>[^\0]+)\0+', self.buffer):
            self.position = position + match.end()
            yield match['record']

        if final and (rem := self.buffer):
            self.position += len(rem)
            yield rem


class LogRecord(typing.NamedTuple):

    level: str
    record: typing.Union[dict, str]

    _Loader = SLoader

    _task_log_pattern = re.compile(r'<([1-5])> *(.*)')

    _deserializer_error = ('{conf_path}: unsupported serialization format: {format!r} '
                           f"(select from: {_Loader.__deserializers__})")

    @staticmethod
    def _normalize_level(value):
        if isinstance(value, int):
            if value < 10:
                value *= 10

            return logging._levelToName.get(value)

        if isinstance(value, str):
            value = value.upper()

            return value if value in logging._nameToLevel else None

        raise TypeError("log level to normalize may be of type int or str "
                        f"not {value.__class__.__name__}")

    @classmethod
    def parse(cls, message: str, format: str, name: str) -> LogRecord:
        if level_match := cls._task_log_pattern.fullmatch(message):
            (record_code, record_text) = level_match.groups()
            record_level = cls._normalize_level(int(record_code))
        else:
            (record_level, record_text) = (None, message)

        try:
            (record_struct, _loader) = cls._Loader.autoload(record_text, format)
        except cls._Loader.NonAutoError:
            try:
                loader = cls._Loader[format]
            except KeyError:
                loader = None

            if loader is None or loader.binary:
                raise ConfValueError(
                    cls._deserializer_error.format(
                        conf_path=f'{name}.format.log',
                        format=format,
                    )
                )

            try:
                record_struct = loader(record_text)
            except loader.raises as exc:
                raise LogRecordDecodeError(
                    format,
                    exc,
                    cls(level=record_level or 'INFO', record=record_text)
                ) from exc

        struct_level = record_struct.get('level') if isinstance(record_struct, dict) else None

        if struct_level:
            if isinstance(struct_level, (int, str)):
                struct_level = cls._normalize_level(struct_level)
            else:
                struct_level = None

        return cls(
            level=record_level or struct_level or 'INFO',
            record=record_struct if isinstance(record_struct, dict) else record_text,
        )



from __future__ import annotations

import abc
import typing

from .. import ext

if typing.TYPE_CHECKING:
    from .event import TaskEvents


class InvokedTask(ext.TaskConfExt):
    """Abstract base to task classes extended for invocation of their
    commands by the operating system.

    """
    @abc.abstractmethod
    def ready_(self) -> bool:
        pass

    @abc.abstractmethod
    def events_(self) -> typing.Optional[TaskEvents]:
        pass



from .base import InvokedTask  # noqa: F401
from .spawned_task import SpawnedTask  # noqa: F401
from .failed_task import (  # noqa: F401
    FailedInvocationTask,
    TaskInvocationError,
)
from .event import (  # noqa: F401
    TaskEvent,
    TaskInvocationFailureEvent,
    TaskLogEvent,
    TaskReadyEvent,
)



from .scheduled_task import ScheduledTask  # noqa: F401
from .invoked_task import (  # noqa: F401
    InvokedTask,
    SpawnedTask,
    FailedInvocationTask,
    TaskEvent,
    TaskInvocationFailureEvent,
    TaskLogEvent,
    TaskReadyEvent,
)



animals = ('aardvark', 'bison', 'canary', 'dalmation', 'emu', 'falcon', 'gnu',
           'hamster', 'impala', 'jellyfish', 'kiwi', 'lemur', 'manatee',
           'nutria', 'okapi', 'porcupine', 'quetzal', 'roadrunner', 'seal',
           'turtle', 'unicorn', 'vole', 'wombat', 'xerus', 'yak', 'zebra')



from __future__ import annotations

import abc
import itertools
import sys
import typing
from dataclasses import dataclass, field, InitVar

from fate.common.log import LogRecord
from fate.common.output import TaskOutput, TaskResult
from fate.conf import OutputEncodeError, ResultEncodeError

if typing.TYPE_CHECKING:
    import datetime

    from fate.util import stream

    from .base import InvokedTask
    from .failed_task import TaskInvocationError


@dataclass
class TaskEvent(abc.ABC):

    task: InvokedTask
    read: bool = field(default=False, init=False)


class TaskInvocationFailureEvent(TaskEvent):

    @property
    def error(self) -> TaskInvocationError:
        return self.task.error


@dataclass
class TaskLogEvent(TaskEvent):

    message: bytes

    def record(self) -> LogRecord:
        return LogRecord.parse(self.message.decode(), self.task.format_['log'], self.task.__name__)


@dataclass
class TaskReadyEvent(TaskEvent):

    returncode: int

    @property
    def duration(self) -> datetime.timedelta:
        return self.task.duration_()

    @property
    def ended(self) -> float:
        return self.task.ended_()

    @property
    def expires(self) -> typing.Optional[float]:
        return self.task.expires_()

    @property
    def stdout(self) -> stream.ProgressiveOutput:
        return self.task.stdout_

    @property
    def stderr(self) -> stream.BufferedOutput:
        return self.task.stderr_

    @property
    def stopped(self) -> typing.Optional[float]:
        return self.task.stopped_

    def results(self) -> typing.List[TaskResult]:
        try:
            outputs = [*TaskOutput.parse(bytes(self.task.stdout_),
                                         self.task.__lib__,
                                         self.task.format_['result'])]
        except OutputEncodeError as exc:
            if result := TaskResult.compose(self.task, exc.output):
                raise ResultEncodeError(result, exc.format, *exc.errors)

            return []
        else:
            return [result
                    for output in outputs
                    if (result := TaskResult.compose(self.task, output))]


@dataclass
class TaskEvents:

    iterable: InitVar[typing.Iterable[TaskEvent]] = ()

    closed: bool = (field(default=False, kw_only=True) if sys.version_info >= (3, 10)
                    else field(default=False))

    _events: typing.List[TaskEvent] = field(default_factory=list, init=False)

    def __post_init__(self, iterable) -> None:
        self._events.extend(iterable)

    def __iter__(self) -> typing.Iterator[TaskEvent]:
        yield from self._events

    def _iter_unread_(self) -> typing.Iterator[TaskEvent]:
        for event in self:
            if not event.read:
                yield event

    def read(self, count=None) -> typing.Iterator[TaskEvent]:
        for event in itertools.islice(self._iter_unread_(), count):
            event.read = True
            yield event

    def write(self, *events: TaskEvent) -> None:
        if self.closed:
            raise ValueError("TaskEvents is closed")

        self._events.extend(events)

    def close(self) -> None:
        self.closed = True



"""Task state persistence via file descriptor input and output."""
import functools

from fate.util.format import Dumper, SLoader


def _ignore_bfd(func):
    """Decorator suppressing exceptions regarding bad file descriptors.

    Should the decorated function raise an `OSError` exception whose
    `errno` is set to `9` -- indicating a bad file descriptor -- the
    exception will be silently suppressed.

    In this case, `None` is returned in lieu of the decorated function's
    typical result (if any).

    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except OSError as exc:
            #
            # errno 9 -- bad file descriptor -- suggests we're just not in the scheduler
            #
            # (that should be OK)
            #
            if exc.errno != 9:
                raise

    return wrapper


@_ignore_bfd
def read(*, format='auto', fd=3):
    """Load (parameterized) state from file descriptor `fd` (defaulting
    to the Fate standard state input descriptor).

    Input is deserialized "auto-magically" according to `format`
    (defaulting to `auto`). The input serialization format may be
    specified as one of: `{}`.

    """
    with open(fd) as file:
        data = file.read()

    try:
        (state, _loader) = SLoader.autoload(data, format)
    except SLoader.NonAutoError:
        pass
    else:
        return state

    try:
        loader = SLoader[format]
    except KeyError:
        raise ValueError(f"unsupported format: {format!r}")

    return loader(data)

read.__doc__ = read.__doc__.format(SLoader.__names__)


@_ignore_bfd
def write(results, /, format='json', fd=4):
    """Write task state to file dsecriptor `fd` (defaulting to the Fate
    standard state output descriptor).

    State is presumed to be structured, *i.e.* in the form of a
    `dict`. These will be serialized according to `format` (defaulting
    to `json`). Supported formats include: `{}`.

    Serialization may be disabled by setting `format` to `None`.

    """
    try:
        dumper = str if format is None else Dumper[format]
    except KeyError:
        raise ValueError(f"unsupported format: {format!r}")

    serialized = dumper(results)

    with open(fd, 'w') as file:
        file.write(serialized)

write.__doc__ = write.__doc__.format(Dumper.__names__)



"""Task-state connector"""
from lmdb_dict import CachedLmdbDict

from fate.conf import ConfValueError, StateEncodeError


class BoundTaskStateManager:
    """Interface to a `TaskStateManager` bound to a specified task.

    All methods will return results serialized for consumption by the
    bound task. Methods do not require specification of the requesting
    task.

    """
    def __init__(self, manager, task):
        self.manager = manager
        self.task = task

    def read(self):
        return self.manager.read(self.task)

    def write(self, output):
        self.manager.write(self.task, output)

    def read_all(self):
        return self.manager.read_all(self.task)


class TaskStateManager:
    """Task state input and output in the configured serialization
    format of the requesting task.

    Values are cached in memory to avoid redundant serializations.

    """
    class _AutoEncodeError(Exception):
        pass

    def __init__(self, state_path):
        self.db = CachedLmdbDict(state_path)

        self._l_cache_ = {}
        self._g_cache_ = {}

    @staticmethod
    def _format_dump(task):
        format_ = task.format_['state']

        if format_ == 'auto':
            return task.format_['param']

        return format_

    @classmethod
    def _dump_state(cls, task, data):
        format_ = cls._format_dump(task)

        try:
            dumper = task._Dumper[format_]
        except KeyError:
            raise ConfValueError(
                f'{task.__name__}: unsupported state serialization format: '
                f"{format_!r} (select from: {task._Dumper.__names__})"
            )
        else:
            return dumper(data)

    @classmethod
    def _load_state(cls, task, output):
        format_ = task.format_['state']

        try:
            (data, loader) = task._Loader.autoload(output, format_)
        except task._Loader.NonAutoError:
            pass
        else:
            if loader is None:
                raise cls._AutoEncodeError

            return data

        try:
            loader = task._Loader[format_]
        except KeyError:
            raise ConfValueError(
                task._serializer_error.format(
                    conf_path=f'{task.__name__}.format.state',
                    format_=format_,
                )
            )

        try:
            return loader(output)
        except loader.raises as exc:
            raise StateEncodeError(format_, exc)

    def bind(self, task):
        return BoundTaskStateManager(self, task)

    def read(self, task):
        """Retrieve task's state in its preferred serialization format.

        Returns None if task has stored no state.

        """
        if task.__name__ not in self._l_cache_:
            data = self.db.get(task.__name__, '')

            if isinstance(data, str):
                serialization = data
            else:
                serialization = self._dump_state(task, data)

            self._l_cache_[task.__name__] = serialization

            return serialization

        return self._l_cache_[task.__name__]

    def write(self, task, output):
        """Persist task's written output state.

        Does nothing if written output is empty or if it does not differ
        from cached state.

        """
        if not output or output == self._l_cache_.get(task.__name__):
            return

        # update caches
        self._g_cache_.clear()
        self._l_cache_[task.__name__] = output

        # deserialize output
        try:
            data = self._load_state(task, output)
        except self._AutoEncodeError:
            data = output

        # update db
        self.db[task.__name__] = data

    def read_all(self, task):
        """Serialize all tasks' states in given task's preferred format.

        """
        format_ = self._format_dump(task)

        if format_ not in self._g_cache_:
            serialization = self._dump_state(task, self.db)

            self._g_cache_[format_] = serialization

            return serialization

        return self._g_cache_[format_]



import os
import time
from dataclasses import dataclass
from pathlib import Path

from descriptors import cachedproperty

from fate.conf import ConfGroup
from fate.util.log import StructLogger


@dataclass
class SchedulerTiming:
    """TaskScheduler timing.

    Features stable (caching) properties for the scheduler's last check,
    current check and when it should next check.

    All such properties return timestamps.

    """
    conf: ConfGroup
    logger: StructLogger
    path_check: Path        # path to empty file with which time of last
                            # check is stored               # noqa: E116

    @cachedproperty
    def time_check(self):
        return time.time()

    @cachedproperty
    def last_check(self):
        return self._check_state_(update=True)

    @property
    def next_check(self):
        return self._next_check_tasks_ or self._next_check_max_

    def _check_state_(self, update=False):
        try:
            stat_result = os.stat(self.path_check)
        except FileNotFoundError:
            last_check = None
        else:
            last_check = stat_result.st_mtime

        if update:
            if not self.path_check.exists():
                self.path_check.touch()

            os.utime(self.path_check, (self.time_check, self.time_check))

        return last_check

    @cachedproperty
    def _next_check_tasks_(self):
        next_check = None

        for task in self.conf.task.values():
            next_check = task.schedule_next_(
                self.time_check,             # t0
                next_check,                  # t1
                next_check,                  # default
                max_years_between_matches=1  # quit if it's that far out
            )

        return next_check

    _next_max_ = 60 * 60 * 24 * 365  # 1 year in seconds

    @property
    def _next_check_max_(self):
        return self.time_check + self._next_max_



import argparse
import functools
import os
import re
import shutil
import subprocess
import sys
import textwrap
import time
import typing

import argcmdr

from fate.common.ext import CompletingTask, TaskConfExt
from fate.common.output import TaskOutput
from fate.conf import OutputEncodeError
from fate.util.term import snip

from .common import CommandInterface


class OneOffExecutor(CommandInterface, argcmdr.Command):
    """Base class for Fate commands that execute tasks directly.

    Subclasses must define `get_command` to specify the task name
    (if any) and command to execute.

    """
    class Command(typing.NamedTuple):

        args: typing.Sequence[str]
        name: str = ''
        stdin: bytes = b''
        format_result: typing.Sequence[str] = ('auto',)
        timeout: typing.Optional[int] = None

    @staticmethod
    def snip_many(iterable, maxlen, cast=str):
        contents = ', '.join(cast(item) for item in iterable)
        if len(contents) > maxlen - 2:
            contents = contents[:maxlen - 6] + ' ...'
        return f'({contents})'

    @staticmethod
    def print_output(name, text):
        """Print report value text formatted appropriately for its
        length (number of lines).

        """
        if '\n' in text:
            print(f'{name}:', textwrap.indent(text.strip(), '  '), sep='\n\n')
        else:
            print(f'{name}:', text)

    @staticmethod
    def present_outputs(outputs):
        for task_output in outputs:
            if task_output.label:
                label = '{}{}'.format(task_output.label, task_output.ext)
            elif task_output.ext:
                label = f'[{task_output.ext}]'
            else:
                label = '<unrecognized>'

            try:
                text = snip(task_output.value.strip().decode(), 300)
            except UnicodeDecodeError:
                text = "<non-text output>"

            if '\n' in text:
                yield label + '\n\n' + textwrap.indent(text, '  ')
            else:
                yield f'{label}: {text}'

    @classmethod
    def print_report(cls, command, retcode, outputs, stderr, error):
        """Print a report of task command execution outcomes."""
        print('Name:', command.name or '-')

        print()

        cls.print_output('Command', ' '.join(command.args))

        print()

        if error:
            if isinstance(error, subprocess.TimeoutExpired):
                status = ('Timeout', f'({error.timeout}s)')
            else:
                raise NotImplementedError(error)
        else:
            status = (cls.CommandStatus.select(retcode), f'(Exit code {retcode})')

        print('Status:', *status)

        print()

        output = '\n\n'.join(cls.present_outputs(outputs))

        cls.print_output('Result', output or '-')

        if stderr:
            print()

            try:
                logs = stderr.decode()
            except UnicodeDecodeError:
                stderr_formatted = "<could not character-decode stderr logs>"
            else:
                # make fate task logging separators -- null byte -- visual
                stderr_formatted = logs.replace('\0', '\n\n').strip() + '\n'

            cls.print_output('Logged (standard error)', stderr_formatted)

    def __init__(self, parser):
        super().__init__(parser)

        parser.add_argument(
            '-o', '--stdout',
            metavar='path',
            type=argparse.FileType('w'),
            help="write command result to path",
        )
        parser.add_argument(
            '-e', '--stderr',
            metavar='path',
            type=argparse.FileType('w'),
            help="write command standard error to path",
        )
        parser.add_argument(
            '--no-report',
            action='store_false',
            dest='report',
            help="do not print command report",
        )

    def __call__(self, args, parser):
        """Execute and report on task command execution."""
        with self.exit_on_error:
            command_spec = self.delegate('get_command')

            if send := getattr(command_spec, 'send', None):
                command = next(command_spec)
            else:
                command = command_spec

            (program, *command_args) = command.args

            executable = shutil.which(program)

            if executable is None:
                hint = ('\nhint: whitespace in program name suggests a misconfiguration'
                        if re.search(r'\s', program) else '')
                parser.exit(127, f'{parser.prog}: error: {program}: '
                                 f'command not found on path{hint}\n')

            try:
                result = subprocess.run(
                    [executable] + command_args,

                    input=command.stdin,

                    timeout=command.timeout,

                    capture_output=True,

                    # it's assumed that even if stdin is set to a TTY it's purposeful
                    # here; so, indicate to task.param.read() not to worry about it:
                    env=dict(os.environ, FATE_READ_TTY_PARAM='1'),
                )
            except subprocess.TimeoutExpired as exc:
                result = None
                returncode = None

                error = exc
                stdout = exc.stdout or b''
                stderr = exc.stderr or b''
            else:
                error = None

                returncode = result.returncode
                stdout = result.stdout
                stderr = result.stderr

            try:
                outputs = (*TaskOutput.parse(stdout, self.conf._lib_, command.format_result),)
            except OutputEncodeError as exc:
                outputs = (exc.output,)

                print("Warn: bad result encoding for configured format",
                      f'{exc.format!r}:',
                      "path suffix ignored:",
                      self.snip_many(exc.errors, 65, repr),
                      file=sys.stderr)
                print()

            if send:
                try:
                    send((result, outputs))
                except StopIteration:
                    pass
                else:
                    raise ValueError("get_command() generated more than one command")

            if args.stdout:
                args.stdout.write(stdout)
                stdout = f'[See {args.stdout.name}]'
                outputs = (TaskOutput(stdout, '<stored>'),)
                args.stdout.close()

            if args.stderr:
                args.stderr.write(stderr)
                stderr = f'[See {args.stderr.name}]'
                args.stderr.close()

            if args.report:
                self.print_report(command, returncode, outputs, stderr, error)

    def get_command(self, args):
        """Determine task name (if any) and command to execute
        from CLI argumentation.

        As a simple method, returns a OneOffExecutor.Command. As a
        generator method, yields only a single element -- the Command --
        and receives the execution result.

        """
        super(argcmdr.Command, self).__call__(args)


"""Decorator to manufacture OneOffExecutor commands from a simple
function defining method `get_command`.

"""
runcmd = functools.partial(argcmdr.cmd, base=OneOffExecutor,
                           binding=True, method_name='get_command')


class CompletedDebugTask(TaskConfExt, CompletingTask):

    @TaskConfExt._constructor_
    def complete(cls, task, *args, **kwargs):
        return cls(task, *args, **kwargs)

    def __init__(self, data, ended=0):
        super().__init__(data)
        self._ended_ = ended or time.time()

    def ended_(self):
        return self._ended_



from __future__ import annotations

import io
import os
import threading
import time
import typing
from dataclasses import dataclass, field

from descriptors import cachedproperty


@dataclass(eq=False)
class BufferedOutput:
    """Buffer of data read from a given file object.

    The descriptor of the given file may have been set blocking or non-
    blocking. By default, it is assumed that this class is given a file
    whose descriptor has been configured for non-blocking reads. In this
    configuration, the `receive` method may be invoked regularly,
    without needlessly blocking execution. Any data that is received may
    be inspected at `data` or by casting the `BufferedOutput` object
    itself to `bytes` or `str`. (Note that this data may be incomplete.)

    Alternatively, a file with a blocking descriptor (the language
    default) may be given. In this case, `receive` will block until the
    file is completely read, (precisely the usual `file.read()`).

    """
    file: typing.BinaryIO
    data: bytes = field(default=b'', init=False)

    def __bytes__(self) -> bytes:
        return self.data

    def __str__(self) -> str:
        return self.data.decode()

    def __iadd__(self, chunk) -> BufferedOutput:
        self.data += chunk
        return self

    def receive(self) -> None:
        # data may be empty/None so we test it
        if read := self.file.read():
            self += read

    def close(self) -> None:
        self.file.close()


@dataclass(eq=False)
class StagedOutput(BufferedOutput):
    """High-performance buffer of data read from a given file object.

    StagedOutput operates like BufferedOutput, with the distinction that
    data is initially "staged", (in an internal list), for improved
    performance. Upon close, this staged data is gathered into user-
    readable data, (available by casting the object to str or bytes).

    The descriptor of the given file object is presumed to be non-
    blocking. This implementation is only necessary when performing very
    large numbers of repeated read operations.

    See: `ProgressiveOutput`.

    """
    _stage: list = field(default_factory=list, init=False)

    def __iadd__(self, chunk) -> StagedOutput:
        self._stage.append(chunk)
        return self

    def close(self) -> None:
        super().close()
        self.data += b''.join(self._stage)
        self._stage.clear()


class ProgressiveOutput(StagedOutput, threading.Thread):
    """Buffer of data which may be read from the given file object in a
    parallel thread.

    As a StagedOutput, the descriptor of the given file object is
    presumed to have been set non-blocking. A new daemon thread may be
    launched via the `start` method, which will (repeatedly) read the
    given file and store its output.

    Read data may be made available for inspection, and the read file
    closed, via the `stop` (alias `close`) method.

    """
    def __init__(self, file: typing.BinaryIO, thread_name: str | None):
        super().__init__(file)
        threading.Thread.__init__(self, name=thread_name, daemon=True)
        self._closed = threading.Event()

    def run(self):
        while not self._closed.is_set():
            time.sleep(1e-6)
            self.receive()

    def close(self):
        self._closed.set()
        self.join()
        super().close()

    stop = close


@dataclass(eq=False)
class BufferedInput:
    """Buffer of data which is written to a given file object in chunks.

    The descriptor of the given file may have been set blocking or non-
    blocking. By default, it is assumed that this class is given a file
    whose descriptor has been configured for non-blocking writes. In
    this configuration, the `send` method may be invoked regularly,
    without needlessly blocking execution.

    """
    data: bytes
    file: typing.BinaryIO
    buffersize: int = io.DEFAULT_BUFFER_SIZE
    position: int = field(default=0, init=False)

    @cachedproperty
    def datasize(self):
        return len(self.data)

    @property
    def finished(self) -> bool:
        return self.file.closed

    def send(self) -> None:
        if self.finished:
            return

        chunk = self.data[self.position:self.buffersize]
        self.file.write(chunk)

        self.position = min(self.position + self.buffersize, self.datasize)

        if self.position == self.datasize:
            try:
                self.file.close()
            except BrokenPipeError:
                pass


def progressive_output(file: typing.BinaryIO, name: str | None) -> ProgressiveOutput:
    """Launch a `ProgressiveOutput` reader in a parallel thread."""
    os.set_blocking(file.fileno(), False)
    reader = ProgressiveOutput(file, name)
    reader.start()
    return reader


def nonblocking_output(file: typing.BinaryIO) -> BufferedOutput:
    """Construct a `BufferedOutput` non-blocking reader of the given
    file.

    The descriptor of the given file is set non-blocking.

    """
    os.set_blocking(file.fileno(), False)
    return BufferedOutput(file)


def nonblocking_input(file: typing.BinaryIO, data: bytes) -> BufferedInput:
    """Construct a `BufferedInput` non-blocking writer of the `data` to
    `file`.

    The descriptor of the given file is set non-blocking.

    """
    os.set_blocking(file.fileno(), False)
    return BufferedInput(data, file)



import argcmdr
import argparse
import os.path
import sys

from fate.common.output import TaskOutput, TaskResult

from .. import CompletedDebugTask, Main, runcmd


READABLE = argparse.FileType('rb')


def path_to_bytes(value: str) -> bytes:
    """Guess whether to use given value as-is or to treat as a
    filesystem path (from which to read a value).

    Returns bytes derived either from the given value OR the result of
    reading a file at that path (via `FileType`).

    """
    if not value.startswith('{') and '\n' not in value and (
        value == '-' or os.path.sep in value or os.path.exists(value)
    ):
        with READABLE(value) as file:
            return file.read()

    return value.encode()


@Main.register
class Debug(argcmdr.Command):
    """ad-hoc execution commands"""

    @runcmd('arguments', metavar='command-arguments', nargs=argparse.REMAINDER,
            help="command arguments (optional)")
    @runcmd('command', help="program to execute")
    @runcmd('-i', '--stdin', metavar='path|text', type=path_to_bytes,
            help="set standard input (parameterization) for command to given "
                 "path or text (specify '-' to pass through stdin)")
    @runcmd('--format-result', default='auto', choices=TaskOutput.modes,
            help="result format to detect (default: %(default)s)")
    def execute(context, args):
        """execute an arbitrary program as an ad-hoc task"""
        return context.Command(
            [args.command] + args.arguments,
            stdin=args.stdin or b'',
            format_result=args.format_result,
        )

    @runcmd('task', help="name of configured task to run")
    @runcmd('-i', '--stdin', metavar='path|text', type=path_to_bytes,
            help="override standard input (parameterization) for task to given "
                 "path or text (specify '-' to pass through stdin) "
                 "(default: from configuration)")
    @runcmd('--record', action='store_true', help="record task result")
    def run(context, args, parser):
        """run a configured task ad-hoc"""
        try:
            spec = context.conf.task[args.task]
        except KeyError:
            parser.error(f"task not found: '{args.task}'")

        (proc, outputs) = yield context.Command(
            args=spec.exec_,
            name=args.task,
            stdin=spec.param_.encode() if args.stdin is None else args.stdin,
            format_result=spec.format_['result'],
            timeout=(timeout := spec.timeout_) and timeout.total_seconds(),
        )

        if outputs and args.record and proc.returncode == 0:
            completed_task = CompletedDebugTask.complete(spec)

            for output in outputs:
                if result := TaskResult.compose(completed_task, output):
                    try:
                        context.write_result(result.path, result.value)
                    except NotADirectoryError as exc:
                        print('cannot record result: path or sub-path is not a directory:',
                              exc.filename,
                              file=sys.stderr)



"""Extension of task configuration with specific functionality."""
from fate.common.ext import TaskConfExt


class BoundTask(TaskConfExt):
    """Abstract base to classes binding a task to its persisted state."""

    def __init__(self, data, /, state):
        super().__init__(data)

        self._state_ = state



import abc
import functools

import descriptors

from fate.conf import types


class CompletingTask(abc.ABC):

    @property
    @abc.abstractmethod
    def path_(self):
        pass

    @abc.abstractmethod
    def ended_(self):
        pass


class TaskConfExt(types.TaskConfDict):
    """Abstract base to classes extending task configuration with
    specific functionality.

    Concrete subclasses must define at least one custom constructor, a
    method which is decorated with TaskConfExt._constructor_. This
    method need only return an instance of the subclass. The method will
    be wrapped as a class-only method and with functionality to ensure
    the constructed instance operates correctly as task configuration.

    """
    _constructors_ = frozenset()

    @staticmethod
    def _link_(extension, task):
        task.__parent__.__adopt__(task.__name__, extension)

    @classmethod
    def _constructor_(bcs, func):
        @functools.wraps(func)
        def wrapped(cls, task, *args, **kwargs):
            instance = func(cls, task, *args, **kwargs)

            if instance.__depth__ is None:
                # force-link the new instance to its parent
                bcs._link_(instance, task)

            return instance

        wrapped._isconstructor_ = True

        return descriptors.classonlymethod(wrapped)

    def __init_subclass__(cls, /, **kwargs):
        super().__init_subclass__(**kwargs)

        cls._constructors_ |= frozenset(
            name for (name, member) in cls.__dict__.items()
            if getattr(getattr(member, '__func__', member), '_isconstructor_', False)
        )

    def __init__(self, *args, **kwargs):
        if not self._constructors_:
            raise TypeError(f"Can't instantiate abstract class {self.__class__.__name__} "
                            "without task adoption-handling constructor(s)")

        super().__init__(*args, **kwargs)

    def __adopt_parent__(self, name, mapping):
        if mapping.__parent__ is None:
            mapping.__parent__ = self
            return

        # we've likely taken over for existing configuration via constructor.
        # rather than insist that child is in our tree, merely check
        # that its tree looks the same.
        assert isinstance(mapping.__parent__, types.TaskConfDict)
        assert mapping.__parent__.__path__ == self.__path__



from __future__ import annotations

import collections
import datetime
import enum
import io
import pathlib
import tarfile
import typing

from descriptors import classonlymethod

from fate.common.ext import CompletingTask
from fate.conf.error import ConfValueError, OutputEncodeError
from fate.util.format import SLoader
from fate.util.iteration import returnboth


class TaskOutput(typing.NamedTuple):

    value: bytes
    label: str = ''
    ext: str = ''

    Loader = SLoader

    class ArchiveMode(enum.Enum):

        file = enum.auto()
        archive = enum.auto()
        detect = enum.auto()

    modes = (
        *Loader.__formats__,
        ArchiveMode.archive.name,
    )

    class NonArchiveError(ValueError):
        pass

    @staticmethod
    def _split_suffixes(name: str, sep: str = '.') -> typing.Tuple[str]:
        """Split file basename into tuple of name and full extension.

        Unlike os.path.splitext, the entirety of the extension is split,
        from the first incidence of the separator.

        """
        if name.endswith(sep):
            return (name, '')

        (head, target) = (name[0], name[1:]) if name.startswith(sep) else ('', name)

        (body, tail_sep, tail) = target.partition(sep)

        return (head + body, tail_sep + tail)

    @classonlymethod
    def iter_archive(cls,
                     stdout: bytes,
                     archive_marker: str,
                     archive_mode: ArchiveMode) -> typing.Iterator[TaskOutput]:
        if not isinstance(archive_mode, cls.ArchiveMode):
            raise TypeError(f"expected mode of type {cls.__name__}.{cls.ArchiveMode.__name__} "
                            f"not {archive_mode.__class__.__name__}")

        if archive_mode is cls.ArchiveMode.file:
            raise ValueError(f"cannot expand archive in non-archive mode {archive_mode}")

        st_file = io.BytesIO(stdout)

        try:
            archive = tarfile.open(fileobj=st_file)
        except tarfile.TarError:
            raise cls.NonArchiveError("payload not recognizable as tar archive")

        if archive_mode is cls.ArchiveMode.detect:
            try:
                marker = archive.getmember(f'.{archive_marker}')
            except KeyError:
                raise cls.NonArchiveError(f"archive marker '.{archive_marker}' "
                                          f"not detected within tar archive")

            if not marker.isdir():
                raise cls.NonArchiveError(
                    f"tar member found with name of archive marker but should "
                    f"be directory (type {tarfile.DIRTYPE}) not type {marker.type}"
                )

        for member in archive:
            if member.isfile():
                yield cls(archive.extractfile(member).read(),
                          *cls._split_suffixes(member.name))

    @classonlymethod
    @returnboth
    def detect_format(
        cls,
        stdout: bytes,
        formats: typing.Iterable[str],
    ) -> typing.Iterator[Exception]:
        formats = set(formats)

        if unsupported_formats := formats.difference(cls.modes):
            raise ConfValueError(f"unexpected format mode(s) {unsupported_formats!r} "
                                 f"not member(s) of {cls.modes!r}")

        try:
            (_struct, loader) = cls.Loader.autoload(stdout, formats)
        except cls.Loader.NonAutoError:
            pass
        else:
            return loader.suffix if loader else ''

        binary = stdout

        try:
            text = binary.decode()
        except UnicodeDecodeError as exc:
            decode_exc = exc
            text = None
        else:
            decode_exc = None

        for format_ in formats:
            try:
                loader = cls.Loader[format_]
            except KeyError:
                raise ConfValueError(f"unexpected format {format_!r} not a member of "
                                     f"{set(cls.Loader.__members__)!r}")

            encoded = binary if loader.binary else text

            if encoded is None:
                yield decode_exc
                continue

            try:
                decoding = loader(encoded)
            except loader.raises as exc:
                yield exc
            else:
                if isinstance(decoding, cls.Loader.Decoding):
                    loader = decoding.decoder

                return loader.suffix

        return ''

    @classonlymethod
    def parse(
        cls,
        stdout: bytes,
        archive_marker: str,
        mode: typing.Union[str, typing.Iterable[str]] = 'auto',
    ) -> typing.Iterator[TaskOutput]:
        if not mode:
            modes = set()
        elif isinstance(mode, str) or not isinstance(mode, collections.abc.Iterable):
            modes = {mode}
        else:
            modes = set(mode)

        if 'archive' in modes:
            archive_mode = cls.ArchiveMode.archive
            modes.remove('archive')
        elif modes & {'auto', 'mixed'}:
            archive_mode = cls.ArchiveMode.detect
        else:
            archive_mode = cls.ArchiveMode.file

        if 'auto' in modes and not stdout:
            # auto (unlike mixed) suppresses empty results: quit
            return

        if archive_mode is not cls.ArchiveMode.file:
            # if not configured for tar output then check for expandable tar archive
            try:
                yield from cls.iter_archive(stdout, archive_marker, archive_mode)
            except cls.NonArchiveError:
                pass
            else:
                return

        (ext, errors) = cls.detect_format(stdout, modes)

        if errors and not ext:
            raise OutputEncodeError(cls(stdout), mode, *errors)

        yield cls(stdout, ext=ext)


class TaskResult(typing.NamedTuple):

    value: bytes
    path: pathlib.Path

    @classonlymethod
    def compose(cls, task: CompletingTask, output: TaskOutput) -> typing.Optional[TaskResult]:
        path = task.path_.result_(
            ext=output.ext,
            label=output.label,
            label_ext=output.label and f'-{output.label}',
            task_ended_at=datetime.datetime.fromtimestamp(task.ended_()),
            task_name=task.__name__,
        )

        return cls(
            value=output.value,
            path=pathlib.Path(path),
        ) if path else None



import abc
import hashlib
import os
import typing

from descriptors import cachedproperty, classproperty

from fate.conf import ConfBracketError
from fate.util.animals import animals
from fate.util.iteration import storeresult

from .state import TaskStateManager
from .task import ScheduledTask
from .timing import SchedulerTiming
from .util.reset import resets, Resets


class TaskScheduler(Resets):
    """Abstract base class of task schedulers.

    Method `exec_tasks` must be implemented by concrete subclasses -- to
    execute tasks -- as a synchronous generator of completed tasks.
    (Task execution may be and likely should be asynchronous / pooled.)

    Invocation of a scheduler instance -- via `__call__` -- returns an
    iterator of completed tasks (derived from `exec_tasks`). Upon
    exhaustion, this iterator's attribute `info` is set to an instance
    of SchedInfo.

    Method `collect_tasks` generates tasks to be executed as a stream of
    ScheduledTask objects.

    Invocation of `collect_tasks` records the current timestamp on the
    file system and under scheduler property `timing.time_check`. The
    timestamp of the previous "check" (retrieved from the file system)
    is stored under scheduler property `timing.last_check`. (See
    `SchedulerTiming`.)

    The Boolean argument `reset` of methods `collect_tasks`, `__call__`,
    (and `reset`), serve to clear the above scheduler properties, such
    that a new "check" -- and a new round of task-collection -- may take
    place. Otherwise, `collect_tasks` is idempotent.

    """
    class SchedInfo(typing.NamedTuple):

        completed_count: int
        next_time: float

    @classproperty
    def module_short(cls):
        (*_root, name) = cls.__module__.rsplit('.', 1)
        return name.replace('_', '-')

    def __init__(self, conf, logger):
        super().__init__()
        self.conf = conf
        self.logger = logger.set(sched=self.module_short)

    @abc.abstractmethod
    def exec_tasks(self, reset=False):
        yield from ()
        return 0

    @storeresult('info')
    def __call__(self, reset=False):
        count = yield from self.exec_tasks(reset=reset)
        return self.SchedInfo(count, self.timing.next_check)

    @resets
    @cachedproperty
    def timing(self):
        return SchedulerTiming(
            self.conf,
            self.logger,
            self.path_state / 'check',
        )

    @cachedproperty
    def state(self):
        return TaskStateManager(self.path_state / 'state')

    @cachedproperty
    def path_state(self):
        """Path to persistent state directory dedicated to the set of
        configuration files of this run.

        The state directory is ensured to be unique for each set of
        configuration file paths, incorporating an MD5 hash in its base
        name for this purpose.

        For the sake of user-friendliness, (or debugger-friendliness),
        a friendly name tag is prepended to this base name, as well.

        For example:

            /var/lib/fate/jellyfish-14a770d5653139e0bb0e39eaf2ec1b67/

        Note that name tags *may* collide; however, full paths should
        not for state directories initialized for unequal sets of
        configuration.

        The set of name tags available *are permitted* to change over
        time -- these are ephemeral aides. Only the hash component of
        the directory name is used to identify it. Should a state
        directory's tag component diverge from the library's
        expectation, it will be renamed appropriately.

        Besides any data written by dependent components, all state
        directories are initialized to contain the subdirectory `conf/`.
        This subdirectory is non-functional and for the purpose of
        debugging. It will contain symbolic links to the configuration
        files of the runs for which the state directory was created.

        """
        # compute conf path-based signature and hash
        signature = os.pathsep.join(sorted(str(conf.__path__) for conf in self.conf))
        file_hash = hashlib.md5(signature.encode()).hexdigest()

        # add in deterministic (but non-unique) friendly name
        name_index = int(file_hash, 16) % len(animals)
        name = animals[name_index]

        # and you've got a friendly, unique path!
        path_state = self.conf._prefix_.state / f"{name}-{file_hash}"

        # check for existing paths with a stale friendly name tag
        if not path_state.exists():
            candidates = (
                (path, path.name.rsplit('-', 1)[-1])
                for path in self.conf._prefix_.state.iterdir()
                if path.is_dir()
            )
            matches = (path for (path, file_hash1) in candidates if file_hash1 == file_hash)

            try:
                (collision, *extras) = matches
            except (FileNotFoundError, ValueError):
                #
                # either the base state dir doesn't yet exist
                # or it does but there are no collisions
                #
                # either way all good: let's initialize the new one
                #
                path_conf = path_state / 'conf'
                path_conf.mkdir(parents=True)

                for conf in self.conf:
                    link_path = path_conf / conf.__path__.name
                    link_path.symlink_to(conf.__path__)
            else:
                # found one: migrate it
                self.logger.debug(
                    stale=str(collision),
                    msg='migrating stale state directory',
                )
                collision.rename(path_state)

                if extras:
                    # wuh oh found more
                    self.logger.warning(
                        stale=[str(path) for path in extras],
                        msg='ignoring additional stale state directories',
                    )

        return path_state

    def _iter_tasks_due_(self):
        # bring last_check & time_check local to ensure consistent generation
        # across resets (as unlikely/discouraged as this situation might be).
        if (last_check := self.timing.last_check) is None:
            return

        time_check = self.timing.time_check

        for task in self.conf.task.values():
            if task.scheduled_(last_check, time_check):
                try:
                    may_schedule = task.if_
                except ConfBracketError as exc:
                    self.logger.warning(task=task.__name__, key=exc.path, msg=str(exc))
                    may_schedule = exc.evaluation

                if not may_schedule:
                    self.logger.info(task=task.__name__,
                                     msg='skipped: suppressed by if/unless condition')
                    continue

                yield ScheduledTask.schedule(task, state=self.state.bind(task))

    def collect_tasks(self, reset=False):
        """Generate ScheduledTasks to be executed.

        If `reset` then the scheduler's current "check" is cleared, such
        that subsequently-scheduled tasks may be generated. Otherwise,
        the set of scheduled tasks is idempotent.

        """
        if reset:
            self.reset()

        yield from self._iter_tasks_due_()



"""Configurable support for serialization formats."""
import bz2
import collections
import csv
import gzip
import io
import json
import lzma
import tarfile
import typing
from functools import partial

import toml
import yaml
from descriptors import classproperty

from fate.util.datastructure import CallableEnum, FileFormatEnum


class ConfigurableYamlLoader(yaml.SafeLoader):
    """YAML SafeLoader supporting alternate collection classes.

    Mapping objects default to the built-in `dict`. Sequences to the
    built-in `list`. Alternate constructors may be specified upon
    initialization of the loader.

    Note: To use with the default interface, `yaml.load()`, a `partial`
    must be constructed, for example:

        yaml.load(stream, functools.partial(ConfigurableYamlLoader,
                                            dict_=NewDict))

    """
    def __init__(self, stream, dict_=None, list_=None):
        super().__init__(stream)
        self.dict_ = dict_
        self.list_ = list_

    def construct_yaml_seq(self, node):
        data = [] if self.list_ is None else self.list_()
        yield data
        data.extend(self.construct_sequence(node))

    def construct_yaml_map(self, node):
        data = {} if self.dict_ is None else self.dict_()
        yield data
        value = self.construct_mapping(node)
        data.update(value)


ConfigurableYamlLoader.add_constructor('tag:yaml.org,2002:seq',
                                       ConfigurableYamlLoader.construct_yaml_seq)

ConfigurableYamlLoader.add_constructor('tag:yaml.org,2002:map',
                                       ConfigurableYamlLoader.construct_yaml_map)


class ConfigurableTomlDecoder(toml.TomlDecoder):

    def __init__(self, _dict=dict, _list=list):
        super().__init__(_dict=_dict)
        self._list = _list

    def load_array(self, a):
        result = super().load_array(a)
        return result if type(result) == self._list else self._list(result)


class JSONEncoder(json.JSONEncoder):
    """JSON encoder supporting abstract Sequence and Mapping objects."""

    def default(self, obj):
        if isinstance(obj, collections.abc.Mapping):
            return dict(obj)

        if isinstance(obj, collections.abc.Sequence):
            return tuple(obj)

        # let base class raise TypeError
        return super().default(obj)

json_encoder = JSONEncoder()


class tag:
    """configurable decorator to set given attribute to given value"""

    _undefined_ = object()

    def __init__(self, name, value=_undefined_):
        self.name = name
        self.value = value

    def __call__(self, value_or_target):
        if self.value is self._undefined_:
            return self.__class__(self.name, value_or_target)

        setattr(value_or_target, self.name, self.value)
        return value_or_target

    def __repr__(self):
        value = 'Undefined' if self.value is self._undefined_ else repr(self.value)
        return f"{self.__class__.__name__}({self.name!r}, {value})"


raises = tag('raises')

auto = tag('auto', True)

binary = tag('binary', True)


class _NameList:

    @classproperty
    def __names__(cls):
        return [member.name for member in cls]


class _Raises:

    @property
    def raises(self):
        return getattr(self.value, 'raises', ())


class Decoder(typing.NamedTuple):

    loader: typing.Callable[[typing.Union[str, bytes]], typing.Any]
    binary: bool
    raises: typing.Tuple[Exception]
    suffix: str

    def __call__(self, value: typing.Union[str, bytes], *args, **kwargs):
        return self.loader(value, *args, **kwargs)


class Decoding(typing.NamedTuple):

    decoded: typing.Any
    decoder: Decoder


def taropen(binary: bytes) -> tarfile.TarFile:
    return tarfile.open(fileobj=io.BytesIO(binary))


class SLoader(_NameList, _Raises, FileFormatEnum, CallableEnum):

    @CallableEnum.member
    @raises(csv.Error)
    def csv(text):
        return list(csv.reader(text.splitlines()))

    @CallableEnum.member
    @auto
    @raises(json.decoder.JSONDecodeError)
    def json(text, dict_=None):
        if dict_ is None:
            return json.loads(text)

        return json.loads(text, object_hook=dict_)

    @CallableEnum.membermethod
    @auto
    @binary
    @raises(tarfile.TarError)
    def tar(self, binary):
        archive = taropen(binary)

        if isinstance(archive.fileobj, gzip.GzipFile):
            suffix = '.tar.gz'
        elif isinstance(archive.fileobj, bz2.BZ2File):
            suffix = '.tar.bz2'
        elif isinstance(archive.fileobj, lzma.LZMAFile):
            suffix = '.tar.xz'
        else:
            suffix = self.suffix

        decoder = self.decoder._replace(loader=taropen, suffix=suffix)

        return self.Decoding(archive, decoder)

    @CallableEnum.member
    @auto
    @raises(toml.decoder.TomlDecodeError)
    def toml(text, **types):
        switched = {f'_{name}'.rstrip('_'): value for (name, value) in types.items()}
        decoder = ConfigurableTomlDecoder(**switched)
        return toml.loads(text, decoder=decoder)

    @CallableEnum.member
    @auto
    @raises(yaml.error.YAMLError)
    def yaml(text, **types):
        conf = yaml.load(text, partial(ConfigurableYamlLoader, **types))
        return {} if conf is None else conf

    @classproperty
    def __deserializers__(cls):
        return [member.name for member in cls if not member.binary]

    @classproperty
    def __auto__(cls):
        return [member for member in cls if member.auto]

    __abstract_formats__ = ('auto', 'mixed')

    @classproperty
    def __formats__(cls):
        return [*cls.__abstract_formats__, *cls.__members__]

    @property
    def auto(self):
        return getattr(self.value, 'auto', False)

    @property
    def binary(self):
        return getattr(self.value, 'binary', False)

    @property
    def decoder(self):
        return self.Decoder(loader=self, binary=self.binary,
                            raises=self.raises, suffix=self.suffix)

    @classmethod
    def autoload(cls, content: typing.Union[str, bytes], format_, **types):
        if not format_:
            return (None, None)

        if isinstance(format_, str) or not isinstance(format_, collections.abc.Iterable):
            formats = {format_}
        else:
            formats = set(format_)

        if formats & {'auto', 'mixed'}:
            if not content:
                return (None, None)

            if isinstance(content, str):
                (binary, text) = (None, content)
            else:
                binary = content

                try:
                    text = binary.decode()
                except UnicodeDecodeError:
                    text = None

            for loader in cls.__auto__:
                encoded = binary if loader.binary else text

                if encoded is None:
                    continue

                try:
                    decoding = loader(encoded, **types)
                except loader.raises:
                    pass
                else:
                    if isinstance(decoding, cls.Decoding):
                        (result, loader) = decoding
                    else:
                        result = decoding

                    # yaml.load treats a string literal as a valid document;
                    # however, this is probably not what's wanted.
                    if not isinstance(result, str):
                        return (result, loader)
            else:
                return (None, None)

        raise cls.NonAutoError(format_)


class NonAutoError(ValueError):
    pass


SLoader.NonAutoError = NonAutoError
SLoader.Decoding = Decoding
SLoader.Decoder = Decoder


class Loader(_Raises, FileFormatEnum, CallableEnum):

    @CallableEnum.member
    @raises(toml.decoder.TomlDecodeError)
    def toml(path, **types):
        with open(path) as fd:
            return SLoader.toml(fd.read(), **types)

    @CallableEnum.member
    @raises(yaml.error.YAMLError)
    def yaml(path, **types):
        with open(path) as fd:
            return SLoader.yaml(fd, **types)


class Dumper(_NameList, CallableEnum):

    @CallableEnum.member
    def json(obj):
        return json_encoder.encode(obj)

    @CallableEnum.member
    def toml(obj):
        return toml.dumps(obj)

    @CallableEnum.member
    def yaml(obj):
        return yaml.dump(obj)



import collections
import functools
import os.path
import pathlib
import re
import typing
import uuid
from datetime import timedelta
from enum import IntEnum

import croniter
import jinja2
from descriptors import classproperty

from fate.util.compat.os import cpu_count
from fate.util.compat.types import NoneType
from fate.util.format import Dumper, SLoader
from fate.util.sentinel import Undefined
from fate.util.timedelta import parse_timedelta

from fate.util.datastructure import (
    adopt,
    at_depth,
    StrEnum,
)

from .. import template

from ..error import (
    ConfBracketError,
    ConfTypeError,
    ConfValueError,
)

from .base import (
    ConfChain,
    ConfDict,
    ConfType,
)


class TaskConfType(ConfType):
    """Generic interface applied to data deserialized from task
    configuration files.

    """
    _Dumper = Dumper
    _Loader = SLoader

    class _DefaultFormat(StrEnum):

        param = 'json'
        log = 'auto'
        result = 'auto'
        state = 'auto'

    class _DefaultScheduling(IntEnum):

        tenancy = (2 * cpu_count() - 1)

    class _ShellExecutable(StrEnum):

        sh = 'sh'
        bash = 'bash'
        python = 'python'

        @classproperty
        def default(cls):
            return cls.sh

        @classproperty
        def options(cls):
            return [str(member) for member in cls]

    class _ScheduleError(StrEnum):

        missing = "{task_name}: schedule requires cron expression like '* * * * *'"
        bad = missing + " not {schedule!r}"
        mixed_hash = missing + (" or, with hashed-name jitter, like 'H H(0-2) * * * H' or, "
                                "with device-unique hashed-name jitter, like 'U U(0-2) * * * U', "
                                "not {schedule!r}")

        def self_format(self, task, schedule=None):
            return self.format(task_name=task.__name__, schedule=schedule)

    _result_format_error = ('{conf_path}: unsupported result format: {format_!r} '
                            f"(select from: {_Loader.__names__})")

    def __repr__(self):
        repr_ = super().__repr__()

        if (path := self.__path__) is None:
            return repr_

        return f"<{self.__class__.__name__} @ {path}: {repr_}>"

    @property
    def __default__(self):
        return self.__root__.__other__.default

    @at_depth(0)
    @property
    @adopt('schedule')
    def schedule_(self):
        try:
            schedule = self['schedule']
        except KeyError:
            raise ConfTypeError(self._ScheduleError.missing.self_format(self))

        if isinstance(schedule, collections.abc.Mapping):
            meta = self.__class__(schedule)

            try:
                schedule = self['schedule']['expression']
            except KeyError:
                raise ConfTypeError(self._ScheduleError.missing.self_format(self))
        else:
            meta = self.__class__(expression=schedule)

        if not isinstance(schedule, str):
            raise ConfTypeError(self._ScheduleError.bad.self_format(self, schedule))

        hash_id = meta.setdefault('hash', self.__name__)

        if not isinstance(hash_id, str):
            raise ConfTypeError(f'{self.__name__}: schedule.hash requires value of type string '
                                f'not {hash_id.__class__.__name__}: {hash_id}')

        return meta

    @at_depth(0)
    def schedule_iter_(self, t0, t1=None, /, *, max_years_between_matches=None):
        meta = self.schedule_

        if meta.expression.startswith('@'):
            (schedule, hash_unique) = (meta.expression, False)
        else:
            (schedule, hash_unique) = re.subn('U', 'H', meta.expression, flags=re.I)

            if hash_unique and re.search('H', meta.expression, re.I):
                raise ConfValueError(
                    self._ScheduleError.mixed_hash.self_format(self, meta.expression)
                )

        hash_id = meta.hash

        if hash_unique:
            hash_id += f'.{uuid.getnode()}'

        # cover for croniter_range not directly supporting hash_id
        croniter_ = functools.partial(croniter.croniter, hash_id=hash_id)

        runs = (croniter_(schedule, t0, max_years_between_matches=max_years_between_matches)
                if t1 is None
                else croniter.croniter_range(t0, t1, schedule, _croniter=croniter_))

        try:
            yield from runs
        except croniter.CroniterBadCronError:
            raise ConfValueError(self._ScheduleError.bad.self_format(self, schedule))

    @at_depth(0)
    def schedule_next_(self, *args, **kwargs):
        arglen = len(args)

        if 1 <= arglen <= 2:
            (times, default) = (args, Undefined)
        elif arglen == 3:
            (*times, default) = args
        else:
            raise TypeError(f"expected 2 or 3 arguments got {arglen}")

        series = self.schedule_iter_(*times, **kwargs)
        return next(series) if default is Undefined else next(series, default)

    @at_depth(0)
    def scheduled_(self, *times):
        try:
            next(self.schedule_iter_(*times))
        except StopIteration:
            return False
        else:
            return True

    @at_depth(0)
    @property
    @adopt('scheduling')
    def scheduling_(self):
        return TaskChainMap(
            self.get('scheduling', {}),
            self.__default__.get('scheduling', {}),
            self._DefaultScheduling.__members__,
        )

    @at_depth(0)
    @property
    def exec_(self) -> typing.Tuple[str]:
        # check for superfluous/conflicting argumentation
        keys = ('exec', 'shell', 'command')
        if sum(key in self for key in keys) > 1:
            raise ConfTypeError(f"{self.__name__}: ambiguous task configuration: "
                                f"specify 0 or 1 of keys: {keys!r}")

        if 'exec' in self:
            try:
                return tuple(template.render_str_list(self['exec']))
            except TypeError:
                raise ConfTypeError(f'{self.__name__}: "exec" requires string or list of strings '
                                    f"not: {self['exec']!r}")
            except jinja2.TemplateError as exc:
                raise ConfValueError(f"{exc.__class__.__name__} @ {self.__path__}.exec: {exc}")

        if 'shell' in self:
            if isinstance(self.shell, collections.abc.Mapping) and 'executable' in self.shell:
                executable = self.shell.executable

                if not isinstance(executable, str):
                    raise ConfTypeError(f'{self.__name__}.shell: "executable" expected '
                                        f'string not: {executable!r}')

                if executable not in self._ShellExecutable:
                    raise ConfValueError(f'{self.__name__}.shell: "executable" must be '
                                         f'one of: {tuple(self._ShellExecutable.options)!r}')
            else:
                executable = self._ShellExecutable.default

            script = (self.shell.get('script') if isinstance(self.shell, collections.abc.Mapping)
                      else self.shell)

            if not isinstance(script, str):
                raise ConfTypeError(
                    f'{self.__name__}: "shell" expected string input '
                    f"for '{executable}' command indicated either by a simple string "
                    f"(for '{self._ShellExecutable.default}') or a mapping of string "
                    f"input ('script') and optional base command "
                    f"('executable': {set(self._ShellExecutable.options)!r})"
                )

            if not script:
                raise ConfValueError(
                    f'{self.__name__}: "shell" expected string input '
                    f"for '{executable}' command indicated either by a simple string "
                    f"(for '{self._ShellExecutable.default}') or a mapping of string "
                    f"input ('script') and optional base command "
                    f"('executable': {set(self._ShellExecutable.options)!r})"
                )

            try:
                return (executable, '-c', template.render_str(script))
            except jinja2.TemplateError as exc:
                raise ConfValueError(f"{exc.__class__.__name__} @ {self.__path__}.shell: {exc}")

        command = self['command'] if 'command' in self else self.__name__

        if not isinstance(command, str):
            raise ConfTypeError(f'{self.__name__}: "command" requires string not: {command!r}')

        return (f'{self.__lib__}-{command}',)

    @at_depth(0)
    @property
    def if_(self):
        if 'if' in self:
            if 'unless' in self:
                raise ConfTypeError(f"{self.__name__}: ambiguous configuration: "
                                    "specify either task 'if' or 'unless' not both")

            (key, negate) = ('if', False)
        elif 'unless' in self:
            (key, negate) = ('unless', True)
        else:
            return True

        expression = self[key]

        if isinstance(expression, bool):
            return not expression if negate else expression

        if not isinstance(expression, str):
            raise ConfTypeError(f'{self.__name__}: "{key}" requires expression '
                                f"string not: {expression!r}")

        bracket_match = template.variable_pattern.fullmatch(expression.strip())

        target = bracket_match['expr'] if bracket_match else expression

        try:
            evaluation = template.eval_expr(target)
        except jinja2.TemplateError as exc:
            raise ConfValueError(f"{exc.__class__.__name__} @ {self.__path__}.{key}: {exc}")

        predicate = not evaluation if negate else bool(evaluation)

        if bracket_match:
            path = self.__get_path__(key)[1:]
            raise ConfBracketError(path, predicate)

        return predicate

    @at_depth(0)
    @property
    def timeout_(self) -> typing.Optional[timedelta]:
        timeout = self.get('timeout', self.__default__.get('timeout'))

        # null, false, '', 0 -> None
        if not timeout:
            return None

        if isinstance(timeout, int):
            # negative, zero -> None
            if timeout <= 0:
                return None

            # positive -> timedelta
            return timedelta(seconds=timeout)

        try:
            # '1w2d3h4m5s' -> timedelta
            return parse_timedelta(timeout)
        except ValueError as exc:
            raise ConfValueError(f'{self.__name__}: timeout: {exc}')
        except TypeError:
            raise ConfTypeError(
                f"{self.__name__}: timeout expects integer (seconds), timedelta string "
                f"(e.g. 1w2d3h4m5s), or falsey/null value, not {timeout.__class__.__name__}"
            )

    @at_depth(0)
    @property
    def format_(self):
        return collections.ChainMap(
            self.get('format', {}),
            self.__default__.get('format', {}),
            self._DefaultFormat.__members__,
        )

    @property
    def _default_path_result(self) -> str:
        return str(
            self._prefix_.data
            / 'result'
            / "result"
              "-{{ task_ended_at.timestamp() | int }}"
              "-{{ task_ended_at.strftime('%Y%m%dT%H%M%S') }}"
              "-{{ task_name }}"
              "{{ label_ext }}"
              "{{ ext }}"
        )

    @at_depth(0)
    @property
    @adopt('path')
    def path_(self):
        return TaskChainMap(
            self.get('path', {}),
            self.__default__.get('path', {}),
            {'result': self._default_path_result},
        )

    @at_depth(0)
    @property
    def param_(self) -> str:
        param = self.get('param', {})

        if isinstance(param, str):
            return param

        format_ = self.format_['param']

        try:
            dumper = self._Dumper[format_]
        except KeyError:
            raise ConfValueError(
                f'{self.__name__}: unsupported serialization format: '
                f"{format_!r} (select from: {self._Dumper.__names__})"
            )
        else:
            return dumper(param)

    @at_depth('*.path')
    def result_(self, **context) -> str:
        spec = self.result

        if not isinstance(spec, (str, NoneType)):
            raise ConfTypeError(f"{self.__path__}.result: expected string or null not "
                                f"{spec.__class__.__name__}")

        # empty path.result synonymous with /dev/null: quit
        if not spec:
            return ''

        default_path = pathlib.Path(
            template.render_template(self._default_path_result, context).strip()
        )

        if not (path := template.render_template(spec, context, default=default_path).strip()):
            return ''

        (path_dir, path_base) = os.path.split(path)

        if path_dir and path_base:
            # all good
            return path

        # fill in blanks with application default
        (default_dir, default_base) = os.path.split(default_path)

        if path_dir:
            return os.path.join(path_dir, default_base)

        return os.path.join(default_dir, path_base)


class TaskConfDict(TaskConfType, ConfDict):
    """Mapping of data deserialized from task configuration files."""


class TaskChainMap(TaskConfType, ConfChain):
    """ChainMap of task configuration with fallback to system-level and
    library-level defaults.

    """



import datetime
import re


TIMEDELTA_PATTERN = re.compile(
    r'(?:(?P<weeks>\d+)w)?'
    r'(?:(?P<days>\d+)d)?'
    r'(?:(?P<hours>\d+)h)?'
    r'(?:(?P<minutes>\d+)m)?'
    r'(?:(?P<seconds>\d+)s)?'
)


def parse_timedelta(text: str) -> datetime.timedelta:
    match = TIMEDELTA_PATTERN.fullmatch(text)

    if not match:
        raise ValueError(f"unsupported timedelta pattern: {text}")

    params = {key: int(value) for (key, value) in match.groupdict().items() if value is not None}

    return datetime.timedelta(**params)


RESOLUTION = datetime.timedelta(milliseconds=1)

MIN = datetime.timedelta(microseconds=1)


def human_readable(delta: datetime.timedelta, resolution: datetime.timedelta = RESOLUTION) -> str:
    # timedelta.__str__ is decent but we can do better (and mirror parse_timedelta)
    #
    # days are fine
    days = delta.days
    #
    # hours and minutes must be extracted from seconds
    (hours, seconds) = divmod(delta.seconds, 3600)
    (minutes, seconds) = divmod(seconds, 60)
    #
    # milliseconds may be extracted from microseconds
    if resolution == MIN:
        (milliseconds, microseconds) = divmod(delta.microseconds, 1000)
    elif resolution == RESOLUTION:
        (milliseconds, microseconds) = (round(delta.microseconds / 1000), 0)
    else:
        raise ValueError(f"unexpected resolution: {resolution}")

    result = ''

    for (value, tag) in (
        (days, 'd'),
        (hours, 'h'),
        (minutes, 'm'),
        (seconds, 's'),
        (milliseconds, 'ms'),
        (microseconds, 'μs'),
    ):
        result += f'{value}{tag}' if value else ''

    if result:
        return result

    # guard against empty (sub-resolution) result
    return '0μs' if resolution == MIN else '0ms'



from .base import InvokedTask
from .event import TaskEvents, TaskInvocationFailureEvent


class TaskInvocationError(LookupError):
    """Exception raised for a task whose command could not be invoked."""


class FailedInvocationTask(InvokedTask):
    """Task whose command could not be invoked."""

    @InvokedTask._constructor_
    def fail(cls, task, err):
        return cls(task, err)

    def __init__(self, data, /, err):
        super().__init__(data)
        self.error = err
        self._events_ = TaskEvents(
            [TaskInvocationFailureEvent(self)],
            closed=True,
        )

    def ready_(self) -> bool:
        return True

    def events_(self) -> TaskEvents:
        return self._events_
